@@ -25,7 +25,7 @@
 #include <pocl_cl.h>
 #include <string.h>
 
-#ifndef _MSC_VER
+#ifndef _WIN32
 #  include <unistd.h>
 #else
 #  include "vccompat.hpp"
@@ -41,6 +41,9 @@ static const char* cpuinfo = "/proc/cpuinfo";
 //Linux' cpufrec interface
 static const char* cpufreq_file="/sys/devices/system/cpu/cpu0/cpufreq/cpuinfo_max_freq";
 
+// Vendor of PCI root bus
+static const char *pci_bus_root_vendor_file = "/sys/bus/pci/devices/0000:00:00.0/vendor";
+
 /* Strings to parse in /proc/cpuinfo. Else branch is for x86, x86_64 */
 #if   defined  __powerpc__
  #define FREQSTRING "clock"
@@ -233,36 +236,6 @@ pocl_cpuinfo_detect_compute_unit_count()
   return -1;  
 }
 
-#ifdef POCL_ANDROID
-
-#define SYSFS_CPU_NUM_CORES_NODE    "/sys/devices/system/cpu/possible"
-
-int
-pocl_sysfs_detect_compute_unit_count()
-{
-  int cores = -1;
-
-  FILE *fp = fopen(SYSFS_CPU_NUM_CORES_NODE, "r");
-
-  // cpu/possible will of format
-  // 0        : for single-core devices
-  // 0-(n-1)  : for n-core cpus
-  if (fp)
-    {
-      cores = fgetc(fp) - '0';
-      if (!feof(fp))         // If more than 1 cores
-        {
-          fgetc(fp);          // Ignore '-'
-          fscanf(fp, "%d", &cores);
-        }
-      fclose(fp);
-      cores ++;             // always printed as (n-1)
-    }
-
-  return cores;
-}
-#endif
-
 #if __arm__ || __aarch64__
 enum
 {
@@ -275,16 +248,19 @@ enum
 
 static const struct
 {
-  unsigned id; /* JEDEC JEP106 code; /proc/cpuinfo, field "CPU implementer" */
+  /* JEDEC JEP106 code; /proc/cpuinfo, field "CPU implementer" */
+  unsigned id;
+  /* PCI vendor ID, to fill the device->vendor_id field */
+  unsigned pci_vendor_id;
   char const *name;
 }
 vendor_list[] =
 {
-  { JEP106_ARM,    "ARM" },
-  { JEP106_BRDCOM, "Broadcom" },
-  { JEP106_CAVIUM, "Cavium" },
-  { JEP106_APM,    "Applied Micro" },
-  { JEP106_QCOM,   "Qualcomm" }
+  { JEP106_ARM,    0x13b5, "ARM" },
+  { JEP106_BRDCOM, 0x14e4, "Broadcom" },
+  { JEP106_CAVIUM, 0x177d, "Cavium" },
+  { JEP106_APM,    0x10e8, "Applied Micro" },
+  { JEP106_QCOM,   0x5143, "Qualcomm" }
 };
 
 typedef struct
@@ -334,6 +310,7 @@ pocl_cpuinfo_get_cpu_name_and_vendor(cl_device_id device)
   /* read contents of /proc/cpuinfo */
   if (access (cpuinfo, R_OK) != 0)
     return;
+
   FILE *f = fopen (cpuinfo, "r");
   char contents[MAX_CPUINFO_SIZE];
   int num_read = fread (contents, 1, MAX_CPUINFO_SIZE - 1, f);            
@@ -345,6 +322,7 @@ pocl_cpuinfo_get_cpu_name_and_vendor(cl_device_id device)
 
 #if __arm__ || __aarch64__
   unsigned vendor_id = -1;
+  size_t i;
 #endif
 #ifdef VENDORSTRING
   do {
@@ -359,10 +337,11 @@ pocl_cpuinfo_get_cpu_name_and_vendor(cl_device_id device)
 #if __arm__ || __aarch64__
     if (1 == sscanf (start, "%x", &vendor_id))
       {
-        for (size_t i = 0; i < sizeof (vendor_list) / sizeof (vendor_list[0]); ++i)
+        for (i = 0; i < sizeof (vendor_list) / sizeof (vendor_list[0]); ++i)
           {
             if (vendor_id == vendor_list[i].id)
               {
+                device->vendor_id = vendor_list[i].pci_vendor_id;
                 start = vendor_list[i].name;
                 end = start + strlen (vendor_list[i].name);
                 break;
@@ -371,7 +350,7 @@ pocl_cpuinfo_get_cpu_name_and_vendor(cl_device_id device)
       }
 #endif
 
-    char *_vendor = malloc(end-start + 1);
+    char *_vendor = (char *)malloc (end - start + 1);
     if (!_vendor)
       break;
     memcpy(_vendor, start, end-start);
@@ -408,7 +387,7 @@ pocl_cpuinfo_get_cpu_name_and_vendor(cl_device_id device)
           break;
       }
 
-      for (size_t i = 0; i < part_count; ++i)
+      for (i = 0; i < part_count; ++i)
         {
           if (part_id == part_list[i].id)
             {
@@ -425,6 +404,25 @@ pocl_cpuinfo_get_cpu_name_and_vendor(cl_device_id device)
   char *new_name = (char*)malloc (len);
   snprintf (new_name, len, "%s-%s", device->short_name, start);
   device->long_name = new_name;
+
+  /* If the vendor_id field is still empty, we should get the PCI ID associated
+   * with the CPU vendor (if there is one), to be ready for the (currently
+   * provisional) OpenCL 3.0 specification that has finally clarified the
+   * meaning of this field. To do this, we look at the vendor advertised by the
+   * PCI root device. At least for Intel and AMD, this does indeed gives us the
+   * expected value. (The alternative would be a look-up table for the vendor
+   * string to the associated PCI ID.)
+   */
+  if (!device->vendor_id)
+    {
+      f = fopen (pci_bus_root_vendor_file, "r");
+      if (f)
+        {
+          /* no error checking, if it failed we just won't have the info */
+          num_read = fscanf (f, "%x", &device->vendor_id);
+          fclose (f);
+        }
+    }
 }
 
 /*
@@ -444,20 +442,29 @@ void
 pocl_cpuinfo_detect_device_info(cl_device_id device) 
 {
   int res;
-#ifdef POCL_ANDROID
-  /* sysfs node seems more suitable for android kernels
-     override the value provided by hwloc
-   */
-  device->max_compute_units = pocl_sysfs_detect_compute_unit_count();
-#else
+
   if (device->max_compute_units == 0) {
     res = pocl_cpuinfo_detect_compute_unit_count();
     device->max_compute_units = (res > 0) ? (cl_uint)res : 0;
   }
-#endif
 
   res = pocl_cpuinfo_detect_max_clock_frequency();
   device->max_clock_frequency = (res > 0) ? (cl_uint)res : 0;
 
   pocl_cpuinfo_get_cpu_name_and_vendor(device);
 }
+
+void
+fix_local_mem_size (cl_device_id device)
+{
+  /* OpenCL 3.0 mandates at least 64KB for CL_DEVICE_MAX_CONSTANT_BUFFER_SIZE
+   * and 32KB for CL_DEVICE_LOCAL_MEM_SIZE. pocl_topology tries to use size of
+   * largest non-shared cache (usually L2), but some CPUs don't have L3
+   * and the only non-shared cache is L1, which can be too small. */
+#if (HOST_DEVICE_CL_VERSION_MAJOR >= 3)
+  if (device->local_mem_size < 32 * 1024)
+    device->local_mem_size = 32 * 1024;
+  if (device->max_constant_buffer_size < 64 * 1024)
+    device->max_constant_buffer_size = 64 * 1024;
+#endif
+}
