@@ -29,18 +29,20 @@ IGNORE_COMPILER_WARNING("-Wstrict-aliasing")
 
 #include "config.h"
 
+#include "clang/Lex/PreprocessorOptions.h"
 #include <clang/Basic/Diagnostic.h>
+#include <clang/Basic/LangOptions.h>
+#include <clang/CodeGen/CodeGenAction.h>
 #include <clang/Driver/Compilation.h>
 #include <clang/Driver/Driver.h>
-#include <clang/CodeGen/CodeGenAction.h>
 #include <clang/Frontend/CompilerInstance.h>
 #include <clang/Frontend/CompilerInvocation.h>
 #include <clang/Frontend/FrontendActions.h>
 #include <clang/Frontend/TextDiagnosticBuffer.h>
 #include <clang/Frontend/TextDiagnosticPrinter.h>
 
-#ifndef LLVM_OLDER_THAN_4_0
-#include "clang/Lex/PreprocessorOptions.h"
+#ifdef LLVM_OLDER_THAN_10_0
+#include "llvm/ADT/ArrayRef.h"
 #endif
 
 #include "llvm/LinkAllPasses.h"
@@ -51,10 +53,13 @@ IGNORE_COMPILER_WARNING("-Wstrict-aliasing")
 #include "llvm/IR/Function.h"
 #include "llvm/IR/Module.h"
 
-#include "llvm/Support/MutexGuard.h"
+#ifndef LLVM_OLDER_THAN_11_0
+#include "llvm/Support/Host.h"
+#endif
 
 #include <iostream>
 #include <sstream>
+#include <regex>
 
 // For some reason including pocl.h before including CodeGenAction.h
 // causes an error. Some kind of macro definition issue. To investigate.
@@ -67,14 +72,13 @@ IGNORE_COMPILER_WARNING("-Wstrict-aliasing")
 #include "pocl_file_util.h"
 #include "pocl_cache.h"
 #include "LLVMUtils.h"
+#include "pocl_util.h"
 
 using namespace clang;
 using namespace llvm;
 
 POP_COMPILER_DIAGS
 
-/* Global pocl device to be used by passes if needed */
-cl_device_id currentPoclDevice = NULL;
 
 
 //#define DEBUG_POCL_LLVM_API
@@ -85,29 +89,6 @@ cl_device_id currentPoclDevice = NULL;
 #endif
 
 
-// Read input source to clang::FrontendOptions.
-// The source is contained in the program->source array,
-// but if debugging option is enabled in the kernel compiler
-// we need to dump the file to disk first for the debugger
-// to find it.
-static inline int
-load_source(FrontendOptions &fe,
-            cl_program program)
-{
-  char source_file[POCL_FILENAME_LENGTH];
-  POCL_RETURN_ERROR_ON(pocl_cache_write_program_source(source_file, program),
-                       CL_OUT_OF_HOST_MEMORY, "Could not write program source");
-
-  fe.Inputs.push_back
-#if LLVM_OLDER_THAN_5_0
-      (FrontendInputFile(source_file, clang::IK_OpenCL));
-#else
-      (FrontendInputFile(source_file, clang::InputKind::OpenCL));
-#endif
-
-  return 0;
-}
-
 // Unlink input sources
 static inline int
 unlink_source(FrontendOptions &fe)
@@ -170,12 +151,11 @@ static void get_build_log(cl_program program,
   appendToProgramBuildLog(program, device_i, log);
 }
 
-static llvm::Module *getKernelLibrary(cl_device_id device);
+static llvm::Module *getKernelLibrary(cl_device_id device,
+                                      PoclLLVMContextData *llvm_ctx);
 
 int pocl_llvm_build_program(cl_program program,
                             unsigned device_i,
-                            const char *user_options_cstr,
-                            char *program_bc_path,
                             cl_uint num_input_headers,
                             const cl_program *input_headers,
                             const char **header_include_names,
@@ -183,15 +163,17 @@ int pocl_llvm_build_program(cl_program program,
 
 {
   char tempfile[POCL_FILENAME_LENGTH];
+  char program_bc_path[POCL_FILENAME_LENGTH];
   tempfile[0] = 0;
-  llvm::Module **mod = NULL;
+  llvm::Module *mod = nullptr;
   char temp_include_dir[POCL_FILENAME_LENGTH];
-  std::string user_options(user_options_cstr ? user_options_cstr : "");
+  std::string user_options(program->compiler_options ? program->compiler_options
+                                                     : "");
   size_t n = 0;
   int error;
-
-  PoclCompilerMutexGuard lockHolder(NULL);
-  InitializeLLVM();
+  cl_context ctx = program->context;
+  PoclLLVMContextData *llvm_ctx = (PoclLLVMContextData *)ctx->llvm_context_data;
+  PoclCompilerMutexGuard lockHolder(&llvm_ctx->Lock);
 
   if (num_input_headers > 0) {
     error = pocl_cache_create_tempdir(temp_include_dir);
@@ -252,29 +234,13 @@ int pocl_llvm_build_program(cl_program program,
       POCL_MEM_FREE(device_switches);
     }
 
-  llvm::StringRef extensions(device->extensions);
+#if !(defined(__x86_64__) && defined(__GNUC__))
+  if (program->flush_denorms) {
+    POCL_MSG_WARN("flush to zero is currently only implemented for "
+                  "x86-64 & gcc/clang, ignoring flag\n");
+  }
+#endif
 
-  std::string cl_ext;
-  if (extensions.size() > 0) {
-    size_t e_start = 0, e_end = 0;
-    while (e_end < std::string::npos) {
-      e_end = extensions.find(' ', e_start);
-      llvm::StringRef tok = extensions.slice(e_start, e_end);
-      e_start = e_end + 1;
-      ss << "-D" << tok.str() << " ";
-#ifndef LLVM_OLDER_THAN_4_0
-      cl_ext += "+";
-      cl_ext += tok.str();
-      cl_ext += ",";
-#endif
-    }
-  }
-#ifndef LLVM_OLDER_THAN_4_0
-  if (!cl_ext.empty()) {
-    cl_ext.back() = ' '; // replace last "," with space
-    ss << "-cl-ext=-all," << cl_ext;
-  }
-#endif
   /* temp dir takes preference */
   if (num_input_headers > 0)
     ss << "-I" << temp_include_dir << " ";
@@ -283,9 +249,7 @@ int pocl_llvm_build_program(cl_program program,
     ss << "-Dcl_khr_int64 ";
 
   ss << "-DPOCL_DEVICE_ADDRESS_BITS=" << device->address_bits << " ";
-#ifndef LLVM_OLDER_THAN_4_0
   ss << "-D__USE_CLANG_OPENCL_C_H ";
-#endif
 
   ss << "-xcl ";
   // Remove the inline keywords to force the user functions
@@ -297,6 +261,36 @@ int pocl_llvm_build_program(cl_program program,
   // required for clGetKernelArgInfo()
   ss << "-cl-kernel-arg-info ";
 
+  size_t fastmath_flag = user_options.find("-cl-fast-relaxed-math");
+
+#if (CLANG_MAJOR == 15)
+#ifdef LLVM_OPAQUE_POINTERS
+  ss << "-opaque-pointers ";
+#else
+  ss << "-no-opaque-pointers ";
+#endif
+#endif
+
+  if (fastmath_flag != std::string::npos) {
+#ifdef ENABLE_CONFORMANCE
+    user_options.replace(fastmath_flag, 21,
+                         "-cl-finite-math-only -cl-unsafe-math-optimizations");
+#endif
+    ss << "-D__FAST_RELAXED_MATH__=1 ";
+  }
+
+#ifdef ENABLE_CONFORMANCE
+  size_t unsafemath_flag = user_options.find("-cl-unsafe-math-optimizations");
+
+  if (unsafemath_flag != std::string::npos) {
+    // this should be almost the same but disables -freciprocal-math.
+    // required for conformance_math_divide test to pass with OpenCL 3.0
+    user_options.replace(unsafemath_flag, 29,
+                         "-cl-no-signed-zeros -cl-mad-enable -ffp-contract=fast");
+  }
+#endif
+
+
   ss << user_options << " ";
 
   if (device->endian_little)
@@ -307,13 +301,11 @@ int pocl_llvm_build_program(cl_program program,
 
   ss << "-DCL_DEVICE_MAX_GLOBAL_VARIABLE_SIZE=" << device->global_var_max_size << " ";
 
-  if (user_options.find("cl-fast-relaxed-math") != std::string::npos)
-    ss << "-D__FAST_RELAXED_MATH__=1 ";
 
-  ss << "-D__OPENCL_VERSION__=" << device->cl_version_int << " ";
+  ss << "-D__OPENCL_VERSION__=" << device->version_as_int << " ";
 
   if (user_options.find("-cl-std=") == std::string::npos)
-    ss << "-cl-std=" << device->cl_version_std << " ";
+    ss << "-cl-std=" << device->opencl_c_version_as_opt << " ";
 
   std::string temp(ss.str());
   size_t pos = temp.find("-cl-std=CL");
@@ -323,6 +315,35 @@ int pocl_llvm_build_program(cl_program program,
   int cl_std_i = cl_std_major * 100 + cl_std_minor * 10;
   ss << "-D__OPENCL_C_VERSION__=" << cl_std_i << " ";
 
+  std::string exts = device->extensions;
+  if (cl_std_major >= 3) {
+    exts += ' ';
+    exts += device->features;
+  }
+  llvm::StringRef extensions(exts);
+
+  std::string cl_ext;
+  if (extensions.size() > 0) {
+    size_t e_start = 0, e_end = 0;
+    while (e_end < std::string::npos) {
+      while (e_start < extensions.size() && std::isspace(extensions[e_start]))
+        ++e_start;
+      if (e_start >= extensions.size())
+        break;
+      e_end = extensions.find(' ', e_start);
+      llvm::StringRef tok = extensions.slice(e_start, e_end);
+      e_start = e_end + 1;
+      ss << "-D" << tok.str() << "=1 ";
+      cl_ext += "+";
+      cl_ext += tok.str();
+      cl_ext += ",";
+    }
+  }
+  if (!cl_ext.empty()) {
+    cl_ext.back() = ' '; // replace last "," with space
+    ss << "-cl-ext=-all," << cl_ext;
+  }
+
   ss << "-fno-builtin ";
   /* with fp-contract=on we get calls to fma with processors which do not
    * have fma instructions. These ruin the performance.
@@ -343,14 +364,45 @@ int pocl_llvm_build_program(cl_program program,
 
   POCL_MSG_PRINT_LLVM("all build options: %s\n", ss.str().c_str());
 
-  std::istream_iterator<std::string> begin(ss);
-  std::istream_iterator<std::string> end;
-  std::istream_iterator<std::string> i = begin;
-  std::vector<const char*> itemcstrs;
+  char WSReplacementChar = 0;
+
+  char *TempOptions = (char *) malloc (ss.str().length() + 1);
+
+  memset (TempOptions, 0, ss.str().length() + 1);
+  strncpy (TempOptions, ss.str().c_str(), ss.str().length());
+
+  if (pocl_escape_quoted_whitespace (TempOptions, &WSReplacementChar) == -1)
+  {
+    POCL_MEM_FREE (TempOptions);
+    return CL_INVALID_BUILD_OPTIONS;
+  }
+
+  std::istringstream iss(TempOptions);
+  std::vector<const char *> itemcstrs;
   std::vector<std::string> itemstrs;
-  while (i != end) {
-    itemstrs.push_back(*i);
-    ++i;
+
+  std::string s;
+
+  while (iss >> s)
+  {
+    // if needed, put back whitespace
+    if (WSReplacementChar != 0)
+    {
+      if (s.find(WSReplacementChar) != std::string::npos)
+      {
+        std::replace(s.begin(), s.end(), WSReplacementChar, ' ');
+      }
+    }
+
+    // if quoted, remove it to make compiler happy
+    if (s.find("\"") != std::string::npos)
+    {
+      std::regex Target("\"");
+      std::string Replacement = " ";
+      s = std::regex_replace(s, Target, Replacement);
+    }
+
+    itemstrs.push_back(s);
   }
 
   for (unsigned idx = 0; idx < itemstrs.size(); idx++) {
@@ -359,6 +411,8 @@ int pocl_llvm_build_program(cl_program program,
     itemcstrs.push_back(itemstrs[idx].c_str());
   }
 
+  POCL_MEM_FREE (TempOptions);
+
 #ifdef DEBUG_POCL_LLVM_API
   // TODO: for some reason the user_options are replicated,
   // they appear twice in a row in the output
@@ -366,13 +420,17 @@ int pocl_llvm_build_program(cl_program program,
             << "user_options: " << user_options << std::endl;
 #endif
 
-  if (program->build_log[device_i])
-    POCL_MEM_FREE(program->build_log[device_i]);
-
-  if (!CompilerInvocation::CreateFromArgs
-      (pocl_build, itemcstrs.data(), itemcstrs.data() + itemcstrs.size(),
-       diags)) {
-    pocl_cache_create_program_cachedir(program, device_i, NULL, 0,
+  if (!CompilerInvocation::CreateFromArgs(
+          pocl_build,
+#ifndef LLVM_OLDER_THAN_10_0
+          ArrayRef<const char *>(itemcstrs.data(),
+                                 itemcstrs.data() + itemcstrs.size()),
+#else
+          itemcstrs.data(), itemcstrs.data() + itemcstrs.size(),
+#endif
+          diags)) {
+    pocl_cache_create_program_cachedir(program, device_i, program->source,
+                                       strlen(program->source),
                                        program_bc_path);
     get_build_log(program, device_i, ss_build_log, diagsBuffer,
                   CI.hasSourceManager() ? &CI.getSourceManager() : nullptr);
@@ -382,17 +440,23 @@ int pocl_llvm_build_program(cl_program program,
   LangOptions *la = pocl_build.getLangOpts();
   PreprocessorOptions &po = pocl_build.getPreprocessorOpts();
 
-#ifdef LLVM_OLDER_THAN_3_9
-  pocl_build.setLangDefaults
-    (*la, clang::IK_OpenCL, clang::LangStandard::lang_opencl12);
+#ifndef LLVM_OLDER_THAN_15_0
+  LangOptions::setLangDefaults(*la, clang::Language::OpenCL, triple,
+                               po.Includes, clang::LangStandard::lang_opencl12);
 #else
-  pocl_build.setLangDefaults
-#if LLVM_OLDER_THAN_5_0
-      (*la, clang::IK_OpenCL, triple, po, clang::LangStandard::lang_opencl12);
+  pocl_build.setLangDefaults(*la,
+#ifndef LLVM_OLDER_THAN_10_0
+                             clang::InputKind(clang::Language::OpenCL),
 #else
-      (*la, clang::InputKind::OpenCL, triple, po,
-       clang::LangStandard::lang_opencl12);
+                             clang::InputKind::OpenCL,
 #endif
+                             triple,
+#ifndef LLVM_OLDER_THAN_12_0
+                             po.Includes,
+#else
+                             po,
+#endif
+                             clang::LangStandard::lang_opencl12);
 #endif
 
   // LLVM 3.3 and older do not set that char is signed which is
@@ -401,11 +465,7 @@ int pocl_llvm_build_program(cl_program program,
 
   // the per-file types don't seem to override this
   la->OpenCLVersion = cl_std_i;
-#ifdef POCL_USE_FAKE_ADDR_SPACE_IDS
-  la->FakeAddressSpaceMap = true;
-#else
   la->FakeAddressSpaceMap = false;
-#endif
   la->Blocks = true; //-fblocks
   la->MathErrno = false; // -fno-math-errno
   la->NoBuiltin = true;  // -fno-builtin
@@ -414,12 +474,17 @@ int pocl_llvm_build_program(cl_program program,
   la->setStackProtector(LangOptions::StackProtectorMode::SSPOff);
 
   la->PICLevel = PICLevel::BigPIC;
+#ifdef __PPC64__
+  la->PIE = 0;
+#else
   la->PIE = 1;
+#endif
 
   std::string IncludeRoot;
   std::string KernelH;
   std::string BuiltinRenamesH;
   std::string PoclTypesH;
+  std::string ClangResourceDir;
 
 #ifdef ENABLE_POCL_BUILDING
   if (pocl_get_bool_option("POCL_BUILDING", 0)) {
@@ -428,7 +493,21 @@ int pocl_llvm_build_program(cl_program program,
   if (0) {
 #endif
   } else {
-    IncludeRoot = POCL_INSTALL_PRIVATE_DATADIR;
+    char temp[POCL_FILENAME_LENGTH];
+    pocl_get_private_datadir(temp);
+    IncludeRoot = temp;
+#ifdef ENABLE_RELOCATION
+    ClangResourceDir = IncludeRoot;
+#endif
+  }
+  if (ClangResourceDir.empty()) {     
+#ifndef LLVM_OLDER_THAN_9_0
+    ClangResourceDir = driver::Driver::GetResourcesPath(CLANG);
+#else
+    DiagnosticsEngine Diags{new DiagnosticIDs, new DiagnosticOptions};
+    driver::Driver TheDriver(CLANG, "", Diags);
+    ClangResourceDir = TheDriver.ResourceDir;
+#endif
   }
   KernelH = IncludeRoot + "/include/_kernel.h";
   BuiltinRenamesH = IncludeRoot + "/include/_builtin_renames.h";
@@ -436,10 +515,11 @@ int pocl_llvm_build_program(cl_program program,
 
   po.Includes.push_back(PoclTypesH);
   po.Includes.push_back(BuiltinRenamesH);
-#ifndef LLVM_OLDER_THAN_4_0
   // Use Clang's opencl-c.h header.
-  po.Includes.push_back(CLANG_RESOURCE_DIR "/include/opencl-c.h");
+#ifndef LLVM_OLDER_THAN_9_0
+  po.Includes.push_back(ClangResourceDir + "/include/opencl-c-base.h");
 #endif
+  po.Includes.push_back(ClangResourceDir + "/include/opencl-c.h");
   po.Includes.push_back(KernelH);
   clang::TargetOptions &ta = pocl_build.getTargetOpts();
   ta.Triple = device->llvm_target_triplet;
@@ -454,26 +534,36 @@ int pocl_llvm_build_program(cl_program program,
   FrontendOptions &fe = pocl_build.getFrontendOpts();
   // The CreateFromArgs created an stdin input which we should remove first.
   fe.Inputs.clear();
-  if (load_source(fe, program) != 0)
-    return CL_OUT_OF_HOST_MEMORY;
+
+  // Read input source to clang::FrontendOptions.
+  // The source is contained in the program->source array,
+  // but if debugging option is enabled in the kernel compiler
+  // we need to dump the file to disk first for the debugger
+  // to find it.
+  char source_file[POCL_FILENAME_LENGTH];
+  POCL_RETURN_ERROR_ON(pocl_cache_write_program_source(source_file, program),
+                       CL_OUT_OF_HOST_MEMORY, "Could not write program source");
+  fe.Inputs.push_back(
+      FrontendInputFile(source_file,
+#ifndef LLVM_OLDER_THAN_10_0
+                        clang::InputKind(clang::Language::OpenCL)
+#else
+                        clang::InputKind::OpenCL
+#endif
+                            ));
 
   CodeGenOptions &cg = pocl_build.getCodeGenOpts();
   cg.EmitOpenCLArgMetadata = true;
   cg.StackRealignment = true;
-  // Let the vectorizer or another optimization pass unroll the loops,
-  // in case it sees beneficial.
-  cg.UnrollLoops = false;
-  // Lets leave vectorization to later compilation phase
-  cg.VectorizeLoop = false;
-  cg.VectorizeSLP = false;
-  // This workarounds a Frontend codegen issues with an illegal address
-  // space cast which is later flattened (and thus implicitly fixed) in
-  // the TargetAddressSpaces. See:  https://github.com/pocl/pocl/issues/195
-#ifdef POCL_USE_FAKE_ADDR_SPACE_IDS
-  cg.VerifyModule = false;
-#else
+  if (!device->spmd) {
+    // Let the vectorizer or another optimization pass unroll the loops,
+    // in case it sees beneficial.
+    cg.UnrollLoops = false;
+    // Lets leave vectorization to later compilation phase
+    cg.VectorizeLoop = false;
+    cg.VectorizeSLP = false;
+  }
   cg.VerifyModule = true;
-#endif
 
   PreprocessorOutputOptions &poo = pocl_build.getPreprocessorOutputOpts();
   poo.ShowCPP = 1;
@@ -483,8 +573,9 @@ int pocl_llvm_build_program(cl_program program,
   poo.ShowMacros = 1;
   poo.RewriteIncludes = 0;
 
-  pocl_cache_tempname(tempfile, ".preproc.cl", NULL);
-  fe.OutputFile.assign(tempfile);
+  error = pocl_cache_tempname(tempfile, ".preproc.cl", NULL);
+  assert(error == 0);
+  fe.OutputFile.assign((const char *)tempfile);
 
   bool success = true;
   clang::PrintPreprocessedAction Preprocess;
@@ -504,7 +595,8 @@ int pocl_llvm_build_program(cl_program program,
   }
 
   if (PreprocessedOut == nullptr) {
-    pocl_cache_create_program_cachedir(program, device_i, NULL, 0,
+    pocl_cache_create_program_cachedir(program, device_i, program->source,
+                                       strlen(program->source),
                                        program_bc_path);
     get_build_log(program, device_i, ss_build_log, diagsBuffer,
                   CI.hasSourceManager() ? &CI.getSourceManager() : nullptr);
@@ -518,14 +610,41 @@ int pocl_llvm_build_program(cl_program program,
 
   unlink_source(fe);
 
-  if (pocl_exists(program_bc_path))
+  if (pocl_exists(program_bc_path)) {
+    char *binary = nullptr;
+    uint64_t fsize;
+    /* Read binaries from program.bc to memory */
+    if (program->binaries[device_i] != nullptr) {
+      POCL_MEM_FREE(program->binaries[device_i]);
+      program->binary_sizes[device_i] = 0;
+    }
+    int r = pocl_read_file(program_bc_path, &binary, &fsize);
+    POCL_RETURN_ERROR_ON(r, CL_BUILD_ERROR,
+                         "Failed to read binaries from program.bc to "
+                         "memory: %s\n",
+                         program_bc_path);
+
+    program->binary_sizes[device_i] = (size_t)fsize;
+    program->binaries[device_i] = (unsigned char *)binary;
+
+    mod = (llvm::Module *)program->data[device_i];
+    if (mod != nullptr) {
+      delete mod;
+      program->data[device_i] = nullptr;
+      --llvm_ctx->number_of_IRs;
+    }
+
+    program->data[device_i] = parseModuleIR(program_bc_path, llvm_ctx->Context);
+    assert(program->data[device_i]);
+    ++llvm_ctx->number_of_IRs;
+
     return CL_SUCCESS;
+  }
 
   // TODO: use pch: it is possible to disable the strict checking for
   // the compilation flags used to compile it and the current translation
   // unit via the preprocessor options directly.
-  llvm::LLVMContext &c = GlobalContext();
-  clang::EmitLLVMOnlyAction EmitLLVM(&c);
+  clang::EmitLLVMOnlyAction EmitLLVM(llvm_ctx->Context);
   success = CI.ExecuteAction(EmitLLVM);
 
   get_build_log(program, device_i, ss_build_log, diagsBuffer, &CI.getSourceManager());
@@ -533,47 +652,50 @@ int pocl_llvm_build_program(cl_program program,
   if (!success)
     return CL_BUILD_PROGRAM_FAILURE;
 
-  mod = (llvm::Module **)&program->llvm_irs[device_i];
-  if (*mod != NULL) {
-    delete *mod;
-    --numberOfIRs;
+  mod = (llvm::Module *)program->data[device_i];
+  if (mod != nullptr) {
+    delete mod;
+    --llvm_ctx->number_of_IRs;
   }
 
-  *mod = EmitLLVM.takeModule().release();
-
-  if (*mod == NULL)
+  mod = EmitLLVM.takeModule().release();
+  if (mod == nullptr)
     return CL_BUILD_PROGRAM_FAILURE;
+  else
+    ++llvm_ctx->number_of_IRs;
 
-  ++numberOfIRs;
-
-  if ((*mod)->getModuleFlag("PIC Level") == nullptr)
-    (*mod)->setPICLevel(PICLevel::BigPIC);
-  if ((*mod)->getModuleFlag("PIE Level") == nullptr)
-    (*mod)->setPIELevel(PIELevel::Large);
+  if (mod->getModuleFlag("PIC Level") == nullptr)
+    mod->setPICLevel(PICLevel::BigPIC);
+#ifndef __PPC64__
+  if (mod->getModuleFlag("PIE Level") == nullptr)
+    mod->setPIELevel(PIELevel::Large);
+#endif
 
   // link w kernel lib, but not if we're called from clCompileProgram()
   // Later this should be replaced with indexed linking of source code
   // and/or bitcode for each kernel.
   if (linking_program) {
-    currentPoclDevice = device;
-    llvm::Module *libmodule = getKernelLibrary(device);
+    llvm::Module *libmodule = getKernelLibrary(device, llvm_ctx);
     assert(libmodule != NULL);
     std::string log("Error(s) while linking: \n");
-    if (link(*mod, libmodule, log)) {
+    if (link(mod, libmodule, log, device->global_as_id,
+             device->device_aux_functions)) {
       appendToProgramBuildLog(program, device_i, log);
-      std::string msg = getDiagString();
+      std::string msg = getDiagString(ctx);
       appendToProgramBuildLog(program, device_i, msg);
-      delete *mod;
-      *mod = nullptr;
-      --numberOfIRs;
+      delete mod;
+      mod = nullptr;
+      --llvm_ctx->number_of_IRs;
       return CL_BUILD_PROGRAM_FAILURE;
     }
   }
 
+  program->data[device_i] = mod;
+
   POCL_MSG_PRINT_LLVM("Writing program.bc to %s.\n", program_bc_path);
 
-  /* Always retain program.bc. Its required in clBuildProgram */
-  error = pocl_write_module(*mod, program_bc_path, 0);
+  /* Always retain program.bc */
+  error = pocl_write_module(mod, program_bc_path, 0);
   if(error)
     return error;
 
@@ -581,7 +703,7 @@ int pocl_llvm_build_program(cl_program program,
    * save program->binaries[i]
    */
   std::string content;
-  writeModuleIR(*mod, content);
+  writeModuleIRtoString(mod, content);
 
   if (program->binaries[device_i])
     POCL_MEM_FREE(program->binaries[device_i]);
@@ -594,151 +716,167 @@ int pocl_llvm_build_program(cl_program program,
   return CL_SUCCESS;
 }
 
-int pocl_llvm_link_program(cl_program program, unsigned device_i,
-                           char *program_bc_path, cl_uint num_input_programs,
-                           unsigned char **cur_device_binaries,
-                           size_t *cur_device_binary_sizes, void **cur_llvm_irs,
-                           int create_library, int spir) {
-
-  std::string concated_binaries;
-  llvm::Module *linked_module = nullptr;
-  size_t n = 0, i;
-  cl_device_id device = program->devices[device_i];
-  llvm::Module **modptr = (llvm::Module **)&program->llvm_irs[device_i];
-  int error;
-
-  currentPoclDevice = device;
-  llvm::Module *libmodule = getKernelLibrary(device);
-  assert(libmodule != NULL);
-
-  PoclCompilerMutexGuard lockHolder(NULL);
-  InitializeLLVM();
-
-  if (spir) {
+/* converts a "spir-unknown-unknown" module triple to target (CPU/CUDA etc)
+ * triple, resets the datalayout to the target datalayout */
+static int pocl_convert_spir_bitcode_to_target(llvm::Module *p,
+                                               llvm::Module *libmodule,
+                                               cl_device_id device) {
 #ifdef ENABLE_SPIR
-
-    POCL_RETURN_ERROR_ON ((device->endian_little == CL_FALSE),
+  const std::string &ModTriple = p->getTargetTriple();
+  if (ModTriple.find("spir") == 0) {
+    POCL_RETURN_ERROR_ON((device->endian_little == CL_FALSE),
                          CL_LINK_PROGRAM_FAILURE,
                          "SPIR is only supported on little-endian devices\n");
+    size_t SpirAddrBits = Triple(ModTriple).isArch64Bit() ? 64 : 32;
 
-    linked_module = parseModuleIRMem((char *)program->binaries[device_i],
-                                     program->binary_sizes[device_i]);
-
-    const std::string &spir_triple = linked_module->getTargetTriple();
-    size_t spir_addrbits = Triple(spir_triple).isArch64Bit() ? 64 : 32;
-
-    if (device->address_bits != spir_addrbits) {
-        delete linked_module;
-        POCL_RETURN_ERROR_ON (1, CL_LINK_PROGRAM_FAILURE,
-                       "Device address bits != SPIR binary triple address "
-                       "bits, device: %s / module: %s\n",
-                       device->llvm_target_triplet, spir_triple.c_str());
+    if (device->address_bits != SpirAddrBits) {
+      delete p;
+      POCL_RETURN_ERROR_ON(1, CL_LINK_PROGRAM_FAILURE,
+                           "Device address bits != SPIR binary triple address "
+                           "bits, device: %s / module: %s\n",
+                           device->llvm_target_triplet, ModTriple.c_str());
     }
 
     /* Note this is a hack to get SPIR working. We'll be linking the
      * host kernel library (plain LLVM IR) to the SPIR program.bc,
      * so LLVM complains about incompatible DataLayouts.
      */
-    linked_module->setTargetTriple(libmodule->getTargetTriple());
-    linked_module->setDataLayout(libmodule->getDataLayout());
-    if (linked_module->getModuleFlag("PIC Level") == nullptr)
-      linked_module->setPICLevel(PICLevel::BigPIC);
-    if (linked_module->getModuleFlag("PIE Level") == nullptr)
-      linked_module->setPIELevel(PIELevel::Large);
+    p->setTargetTriple(libmodule->getTargetTriple());
+    p->setDataLayout(libmodule->getDataLayout());
 
-#else
-    POCL_MSG_ERR("SPIR not supported\n");
-    return CL_LINK_PROGRAM_FAILURE;
+    if (p->getModuleFlag("PIC Level") == nullptr)
+      p->setPICLevel(PICLevel::BigPIC);
+#ifndef __PPC64__
+    if (p->getModuleFlag("PIE Level") == nullptr)
+      p->setPIELevel(PIELevel::Large);
 #endif
-  } else {
-
-#ifdef LLVM_OLDER_THAN_3_8
-  llvm::Module *mod =
-      new llvm::Module(StringRef("linked_program"), GlobalContext());
+  }
+  return CL_SUCCESS;
 #else
+  POCL_MSG_ERR("SPIR not supported\n");
+  return CL_LINK_PROGRAM_FAILURE;
+#endif
+}
+
+int pocl_llvm_link_program(cl_program program, unsigned device_i,
+                           cl_uint num_input_programs,
+                           unsigned char **cur_device_binaries,
+                           size_t *cur_device_binary_sizes, void **cur_llvm_irs,
+                           int link_device_builtin_library,
+                           int linking_into_new_cl_program) {
+
+  char program_bc_path[POCL_FILENAME_LENGTH];
+  std::string concated_binaries;
+  size_t n = 0, i;
+  cl_device_id device = program->devices[device_i];
+  llvm::Module **modptr = (llvm::Module **)&program->data[device_i];
+  int error;
+  cl_context ctx = program->context;
+  PoclLLVMContextData *llvm_ctx = (PoclLLVMContextData *)ctx->llvm_context_data;
+  PoclCompilerMutexGuard lockHolder(&llvm_ctx->Lock);
+
+  llvm::Module *libmodule = getKernelLibrary(device, llvm_ctx);
+  assert(libmodule != NULL);
+
   std::unique_ptr<llvm::Module> mod(
-      new llvm::Module(StringRef("linked_program"), GlobalContext()));
-#endif
+      new llvm::Module(StringRef("linked_program"), *llvm_ctx->Context));
+  llvm::Module *LinkedModule = nullptr;
+  std::unique_ptr<llvm::Module> TempModule;
 
+  // link the provided modules together into a single module
   for (i = 0; i < num_input_programs; i++) {
     assert(cur_device_binaries[i]);
     assert(cur_device_binary_sizes[i]);
-    concated_binaries.append(std::string((char *)cur_device_binaries[i],
-                                         cur_device_binary_sizes[i]));
+    concated_binaries.append((char *)cur_device_binaries[i],
+                             cur_device_binary_sizes[i]);
 
-    llvm::Module *p = (llvm::Module *)cur_llvm_irs[i];
-    assert(p);
-
-#ifdef LLVM_OLDER_THAN_3_8
-    if (Linker::LinkModules(mod, llvm::CloneModule(p))) {
-      delete mod;
-#elif LLVM_OLDER_THAN_7_0
-    if (Linker::linkModules(*mod, llvm::CloneModule(p))) {
+    if (cur_llvm_irs && cur_llvm_irs[i]) {
+      llvm::Module *Ptr = (llvm::Module *)cur_llvm_irs[i];
+#ifdef LLVM_OLDER_THAN_7_0
+      TempModule = llvm::CloneModule(Ptr);
 #else
-    if (Linker::linkModules(*mod, llvm::CloneModule(*p))) {
+      TempModule = llvm::CloneModule(*Ptr);
 #endif
-      std::string msg = getDiagString();
+    } else {
+      llvm::Module *Ptr =
+          parseModuleIRMem((char *)cur_device_binaries[0],
+                           cur_device_binary_sizes[0], llvm_ctx->Context);
+      POCL_RETURN_ERROR_ON((Ptr == nullptr), CL_LINK_PROGRAM_FAILURE,
+                           "could not parse module\n");
+      TempModule.reset(Ptr);
+    }
+
+    error = pocl_convert_spir_bitcode_to_target(TempModule.get(), libmodule,
+                                                device);
+    POCL_RETURN_ERROR_ON((error != CL_SUCCESS), CL_LINK_PROGRAM_FAILURE,
+                         "could connvert SPIR to Target\n");
+
+    if (Linker::linkModules(*mod, std::move(TempModule))) {
+      std::string msg = getDiagString(ctx);
       appendToProgramBuildLog(program, device_i, msg);
       return CL_LINK_PROGRAM_FAILURE;
     }
   }
 
-#ifdef LLVM_OLDER_THAN_3_8
-  linked_module = mod;
-#else
-  linked_module = mod.release();
-#endif
-  }
-
-  if (linked_module == nullptr)
+  LinkedModule = mod.release();
+  if (LinkedModule == nullptr)
     return CL_LINK_PROGRAM_FAILURE;
-
+  // delete previous build of program
   if (*modptr != nullptr) {
     delete *modptr;
-    --numberOfIRs;
+    --llvm_ctx->number_of_IRs;
     *modptr = nullptr;
   }
 
-  if (!create_library) {
+  // link the builtin library
+  if (link_device_builtin_library) {
     // linked all the programs together, now link in the kernel library
     std::string log("Error(s) while linking: \n");
-    if (link(linked_module, libmodule, log)) {
+    if (link(LinkedModule, libmodule, log, device->global_as_id,
+             device->device_aux_functions)) {
       appendToProgramBuildLog(program, device_i, log);
-      std::string msg = getDiagString();
+      std::string msg = getDiagString(ctx);
       appendToProgramBuildLog(program, device_i, msg);
-      delete linked_module;
+      delete LinkedModule;
       return CL_BUILD_PROGRAM_FAILURE;
     }
   }
 
-  *modptr = linked_module;
-  ++numberOfIRs;
+  *modptr = LinkedModule;
+  ++llvm_ctx->number_of_IRs;
 
-  /* TODO currently cached on concated binary contents (in undefined order),
-     this is not terribly useful (but we have to store it somewhere..) */
-  error = pocl_cache_create_program_cachedir(program, device_i,
-                                     concated_binaries.c_str(),
-                                     concated_binaries.size(),
-                                     program_bc_path);
-  if (error)
-    {
-      POCL_MSG_ERR ("pocl_cache_create_program_cachedir(%s)"
-                    " failed with %i\n", program_bc_path, error);
+  /* if we're linking binaries into a new cl_program, create cache
+   * on concated binary contents (in undefined order); this is not
+   * terribly useful, but we have to store it somewhere.. */
+  if (linking_into_new_cl_program) {
+    // assert build_hash is empty
+    unsigned bhash_valid = pocl_cache_buildhash_is_valid (program, device_i);
+    assert (!bhash_valid);
+    error = pocl_cache_create_program_cachedir(
+        program, device_i, concated_binaries.c_str(), concated_binaries.size(),
+        program_bc_path);
+    if (error) {
+      POCL_MSG_ERR("pocl_cache_create_program_cachedir(%s)"
+                   " failed with %i\n",
+                   program_bc_path, error);
       return error;
     }
+  } else {
+    /* If we're linking existing cl_program, just get the path
+     * assumes the program->build_hash[i] is already valid. */
+    pocl_cache_program_bc_path(program_bc_path, program, device_i);
+  }
 
   POCL_MSG_PRINT_LLVM("Writing program.bc to %s.\n", program_bc_path);
 
-  /* Always retain program.bc. Its required in clBuildProgram */
-  error = pocl_write_module(linked_module, program_bc_path, 0);
+  /* Always retain program.bc for metadata */
+  error = pocl_write_module(LinkedModule, program_bc_path, 0);
   if (error)
     return error;
 
-  /* To avoid writing & reading the same back,
-   * save program->binaries[i]
-   */
+  /* To avoid writing & reading the same back, save program->binaries[i] */
   std::string content;
-  writeModuleIR(linked_module, content);
+  writeModuleIRtoString(LinkedModule, content);
 
   if (program->binaries[device_i])
     POCL_MEM_FREE(program->binaries[device_i]);
@@ -755,9 +893,15 @@ int pocl_llvm_link_program(cl_program program, unsigned device_i,
 #ifdef KERNELLIB_HOST_DISTRO_VARIANTS
 const char *getX86KernelLibName() {
   StringMap<bool> Features;
-  llvm::sys::getHostCPUFeatures(Features);
   const char *res = NULL;
 
+  if (!llvm::sys::getHostCPUFeatures(Features)) {
+    POCL_MSG_WARN ("getX86KernelLibName(): LLVM can't get host CPU flags!\n");
+    /* getX86KernelLibName should only ever be enabled
+       on x86-64, which always has sse2 */
+    return "sse2";
+  }
+
   if (Features["sse2"])
     res = "sse2";
   else
@@ -785,24 +929,24 @@ const char *getX86KernelLibName() {
 #endif
 
 
-static std::map<cl_device_id, llvm::Module *> kernelLibraryMap;
-
 /**
  * Return the OpenCL C built-in function library bitcode
  * for the given device.
  */
-static llvm::Module* getKernelLibrary(cl_device_id device)
-{
+static llvm::Module *getKernelLibrary(cl_device_id device,
+                                      PoclLLVMContextData *llvm_ctx) {
   Triple triple(device->llvm_target_triplet);
+  llvm::LLVMContext *llvmContext = llvm_ctx->Context;
+  kernelLibraryMapTy *kernelLibraryMap = llvm_ctx->kernelLibraryMap;
 
-  if (kernelLibraryMap.find(device) != kernelLibraryMap.end())
-    return kernelLibraryMap[device];
+  if (kernelLibraryMap->find(device) != kernelLibraryMap->end())
+    return kernelLibraryMap->at(device);
 
   const char *subdir = "host";
   bool is_host = true;
   // TODO: move this to the device layer, a property to ask for
   // the kernel builtin bitcode library name, including its subdir
-#ifdef TCE_AVAILABLE
+#if defined(TCE_AVAILABLE)
   if (triple.getArch() == Triple::tce || triple.getArch() == Triple::tcele) {
     subdir = "tce";
     is_host = false;
@@ -830,6 +974,7 @@ static llvm::Module* getKernelLibrary(cl_device_id device)
 
   std::string kernellib;
   std::string kernellib_fallback;
+
 #ifdef ENABLE_POCL_BUILDING
   if (pocl_get_bool_option("POCL_BUILDING", 0)) {
     kernellib = BUILDDIR;
@@ -837,18 +982,27 @@ static llvm::Module* getKernelLibrary(cl_device_id device)
     kernellib += subdir;
   } else // POCL_BUILDING == 0, use install dir
 #endif
-  kernellib = POCL_INSTALL_PRIVATE_DATADIR;
+  {
+    char temp[POCL_FILENAME_LENGTH];
+    pocl_get_private_datadir(temp);
+    kernellib = temp;
+  }
+
   kernellib += "/kernel-";
   kernellib += device->llvm_target_triplet;
+
   if (is_host) {
     kernellib += '-';
 #ifdef KERNELLIB_HOST_DISTRO_VARIANTS
     kernellib += getX86KernelLibName();
+#elif defined(HOST_CPU_FORCED)
+    kernellib += OCL_KERNEL_TARGET_CPU;
 #else
     kernellib_fallback = kernellib;
     kernellib_fallback += OCL_KERNEL_TARGET_CPU;
     kernellib_fallback += ".bc";
-    kernellib += device->llvm_cpu;
+    if (device->llvm_cpu)
+      kernellib += device->llvm_cpu;
 #endif
   }
   kernellib += ".bc";
@@ -858,36 +1012,27 @@ static llvm::Module* getKernelLibrary(cl_device_id device)
   if (pocl_exists(kernellib.c_str()))
     {
       POCL_MSG_PRINT_LLVM("Using %s as the built-in lib.\n", kernellib.c_str());
-      lib = parseModuleIR(kernellib.c_str());
+      lib = parseModuleIR(kernellib.c_str(), llvmContext);
     }
   else
     {
-#ifndef KERNELLIB_HOST_DISTRO_VARIANTS
+#if !defined(KERNELLIB_HOST_DISTRO_VARIANTS) && !defined(HOST_CPU_FORCED)
       if (is_host && pocl_exists(kernellib_fallback.c_str()))
         {
           POCL_MSG_WARN("Using fallback %s as the built-in lib.\n",
                         kernellib_fallback.c_str());
-          lib = parseModuleIR(kernellib_fallback.c_str());
+          lib = parseModuleIR(kernellib_fallback.c_str(), llvmContext);
         }
       else
 #endif
         POCL_ABORT("Kernel library file %s doesn't exist.\n", kernellib.c_str());
     }
   assert (lib != NULL);
-  kernelLibraryMap[device] = lib;
+  kernelLibraryMap->insert(std::make_pair(device, lib));
 
   return lib;
 }
 
-void cleanKernelLibrary() {
-  for (auto i = kernelLibraryMap.begin(), e = kernelLibraryMap.end();
-       i != e; ++i) {
-    delete (llvm::Module *)i->second;
-  }
-  kernelLibraryMap.clear();
-}
-
-#ifndef LLVM_OLDER_THAN_5_0
 /**
  * Invoke the Clang compiler through its Driver API.
  *
@@ -929,4 +1074,3 @@ int pocl_invoke_clang(cl_device_id Device, const char** Args) {
   }
 
 }
-#endif
